/*** PNG decoder main module */

#include "png.h"

#define NULL ((void*)0)


/** bitXg[錾 */
typedef struct BitBuffer *BitBuffer_t;
static inline void BitBuffer_init(BitBuffer_t buf, byte *p);
static inline uint BitBuffer_getBits(BitBuffer_t buf, int len);

/** memallocate ̂߂̃q[v錾 */
typedef struct Heap *Heap_t;
static inline void Heap_init(Heap_t this, byte *lower, int size);
static void *Heap_alloc(Heap_t this, int size);

/** Chunk 錾 */
typedef struct Chunk *Chunk_t;
static inline uint GETDW(void *p);
static int Chunk_readChunk(Chunk_t this, CRC32_t crc, byte **fpp, byte *end);
static int Chunk_readIhdr(Chunk_t this, Png_t png);
static int Chunk_readPlte(Chunk_t this, Png_t png);
static void Chunk_readIdat(Chunk_t this, Png_t png);

enum CHUNK_TYPE {
  IHDR = 0x49484452,
  PLTE = 0x504C5445,
  IDAT = 0x49444154,
  IEND = 0x49454E44,
  sBIT = 0x73424954,
  tRNS = 0x74524E53,
  pHYs = 0x70485973,
  tEXt = 0x74455874,
};

/** fR[h֌W̊֐錾 */
static inline int ABS(int x);
static int paeth_predictor (int a,int b,int c);
static void scaling(Png_t png, int xs, int ys, int xp, int yp);
static inline byte alphaBlend(byte x, byte y, byte a);


/** q[v\̐錾 */
typedef struct Heap {
  byte *top, *p, *upper;
} Heap;

/** bitXg[\̐錾 */
typedef struct BitBuffer {
	byte *p;
	int cnt;
	byte tmp;
} BitBuffer;

/** Chunk\̐錾 */
typedef struct Chunk {
  byte *p;
  int len;
} Chunk;


/** Chunk ` */
static inline uint GETDW(void *p)
{
  byte *c = p;
  return c[0]<<24 | c[1]<<16 | c[2]<<8 | c[3];
}

static int Chunk_readChunk(Chunk_t this, CRC32_t crc, byte **fpp, byte *end)
{
  uint type, crc1 = 0, crc2;
  byte *fp = *fpp;

  if (fp + 12 <= end) {		/* ̃`N12oCgKv */
    this->len = GETDW(fp);
    if (fp + 12 + this->len <= end) {
      fp += 4;
      type = GETDW(fp);
      if (crc)
	crc1 = CRC32_update(crc, 0, fp, this->len + 4);
      fp += 4;
      this->p = fp;
      fp += this->len;
      if (crc) {
	crc2  = GETDW(fp);
	if (crc1 != crc2)
	  return ERROR_CRC;
      }
      fp += 4;
      *fpp = fp;
    } else
      goto chunkError;
  } else {
  chunkError:
    type = -ERROR_DATA;		/* `N؂Ă */
  }
  return type;
}

static int Chunk_readIhdr(Chunk_t this, Png_t png)
{
  byte filter;
  byte *chunk = this->p;

  png->width = GETDW(chunk),  chunk += 4;
  png->height = GETDW(chunk), chunk += 4;
  png->bit_depth = *chunk++;
  png->clr_type = *chunk++;

  /* J[^Cvƃrbg[x̑gݍ킹`FbN */
  if ((png->clr_type & PALETTE_USED) && png->clr_type != INDEXCOLOR)
    return ERROR_COLORTYPE;
  if ((png->bit_depth & (png->bit_depth - 1)) || png->bit_depth > 16)
    return ERROR_DEPTH;		/* only 1,2,4,8,16 valid */
  if (png->clr_type == INDEXCOLOR) { /* type 3 */
    if (png->bit_depth > 8)
      return ERROR_DEPTH;
  } else if (png->clr_type && png->bit_depth < 8) /* type 2,4,6 */
    return ERROR_DEPTH;

  /* bpp  scan_width ̐ݒ */
  png->bpp = png->bit_depth;	/* bit per pixel for filtering */
  if ((png->clr_type & (TRUECOLOR_USED | PALETTE_USED)) == TRUECOLOR_USED )
    png->bpp *= 3;
  if (png->clr_type & ALPHA_USED)
    png->bpp += png->bit_depth;
  png->scan_width = (png->width * png->bpp + 7) / 8;

  if (*chunk++ != 0)
    return ERROR_COMPRESS;

  filter = *chunk++;
  if (filter != 0)
    return ERROR_FILTER;

  png->interlace = *chunk++;
  if (png->interlace > 1)
    return ERROR_INVALIDINTERLACE;

  return 0;
}

static int Chunk_readPlte(Chunk_t this, Png_t png)
{
  Color *plte = png->plte;
  byte *chunk = this->p;
  int i, nidx;

  if (this->len % 3)
    return ERROR_PALETTE;

  nidx = this->len / 3;
  for (i = 0; i < nidx; i++) {
    plte[i].r = *chunk++;
    plte[i].g = *chunk++;
    plte[i].b = *chunk++;
  }
  return 0;
}

static void Chunk_readIdat(Chunk_t this, Png_t png)
{
  if (! png->zfmt) {
    png->zfmt = this->p;
    png->zlen = this->len;
  } else if (! png->idat) {
    png->idat = this->p - 8;
  }
}

/** q[v` */
static inline void Heap_init(Heap_t this, byte *lower, int size)
{
  this->top = this->p = lower;
  this->upper = lower + size;
}

static void *Heap_alloc(Heap_t this, int size)
{
  byte *p = this->p;
  if (p + size > this->upper)
    return NULL;
  this->p += size;
  return p;
}

/** pngfR[_` */
int Png_parse(Png_t png, byte **fpp, byte *end)
{
  int type;
  CRC32 crc[1];			/* STACK */
  Chunk chunk[1];		/* STACK */
  uint *hp = (uint *)*fpp;

  if (*fpp + 8 <= end) {
    /* char sig[] = { 0x89, 0x50, 0x4e, 0x47, 0x0d, 0x0a, 0x1a, 0x0a }; */
    if (hp[0] != 0x474e5089 || hp[1] != 0x0a1a0a0d)
      goto signatureError;
    *fpp += 8;

    png->idat = png->zfmt = NULL;

    CRC32_init(crc);
    for(;;) {
      type = Chunk_readChunk(chunk, crc, fpp, end);

      switch (type) {
      case IHDR:
	/* ꂪG[ɂȂȂ΂̃`N͓ǂ߂ */
	type = Chunk_readIhdr(chunk, png);
	break;

      case PLTE:
	type = Chunk_readPlte(chunk, png);
	break;

      case IDAT:
	Chunk_readIdat(chunk, png);
	type = 0;
	break;

      case IEND:
	return 0;

      default:
	/* ERROR/Ή̃`N͓ǂݔ΂ */
	break;
      }
      if (0 < type && type < ERROR_END)
	return -type;		/* typeł͂Ȃerror */
    }
  } else {
  signatureError:
    return -ERROR_SIGNATURE;
  }
}

uint Png_getWidth(Png_t png)
{
  return png->width;
}

uint Png_getHeight(Png_t png)
{
  return png->height;
}

uint Png_getRequiredSize(Png_t png)
{
  return sizeof(Inflator) + png->scan_width * 2
    + sizeof(ByteWriter) + sizeof(BitReader);
}

int Png_memallocate(Png_t png, byte *top, uint size, byte *image, Gosakaku_t gosakaku)
{
  Heap heap[1];

  Heap_init(heap, top, size);
  png->inf = Heap_alloc(heap, sizeof(Inflator));
  png->buf = Heap_alloc(heap, sizeof(ByteWriter));
  png->stream = Heap_alloc(heap, sizeof(BitReader));
  png->prev = Heap_alloc(heap, png->scan_width);
  png->line = Heap_alloc(heap, png->scan_width);
  png->gosakaku = gosakaku;
  if (gosakaku) {
    /* true colorobt@ƌ덷obt@ */
    png->true = Heap_alloc(heap, png->width * 4 * 2);
  } else {			/* true color mode */
    png->true = image;
  }
  png->image = image;

  if (! png->inf || ! png->prev || ! png->line || ! png->true
      || ! png->buf || ! png->stream)
    return -1;

  return 0;
}

int Png_decode(Png_t png, int lineLength)
{
  int ret, i;

  png->x = -1;			/* -1filterȂ */
  png->y = 0;
  png->dlen = (png->scan_width + 1) * png->height; /* +1 ̓tB^[ */
  png->imglen = png->width * png->height * 4;
  png->lineLength = lineLength;

  /* 덷obt@NA */
  if (png->gosakaku) {
    int *deltaBuf = NULL;
    if (png->interlace)
      return -ERROR_INTERLACE;
    deltaBuf = (int*)&png->true[png->width * 4];
    for (i = 0; i < png->width; i++)
      deltaBuf[i] = 0;
  } else if (png->interlace) {
    png->pass = 1;
  }

  /* 8rbg̃O[XP[̓pbgőΏ */
  if (png->bit_depth < 8 && ! (png->clr_type & TRUECOLOR_USED)) {
    int m, n;
    m = (1 << png->bit_depth) - 1;
    for (n = 0; n <= m; n++)
      png->plte[n].r = png->plte[n].g = png->plte[n].b = 255 * n / m;
  }

  BitReader_init(png->stream, png->zfmt, png->zlen, png);
  ByteWriter_init(png->buf, png);
  init_inflate(png->inf, png->stream, png->buf);

  ret = decomp_zlib(png->inf);
  if (ret != SZL_OK)
    return ret;

  return 0;
}

static inline int ABS(int x)
{
  return x < 0 ? -x : x;
}

static int paeth_predictor (int a,int b,int c)
{
    int p, pa, pb, pc;    

    p  = a + b - c;
    pa = ABS(p - a);
    pb = ABS(p - b);
    pc = ABS(p - c);

    if ((pa <= pb) && (pa <= pc))
		return a;

    else if (pb <= pc)
		return b;

    else
		return c;
}

/** IDAT`N̒gbitReaderɓn */
byte *Png_getNextIdat(Png_t png, int *size)
{
  if (png->idat) {
    Chunk chunk[1];
    uint type = Chunk_readChunk(chunk, NULL, &png->idat, (void*)0xffffffff);
    if (type == IDAT) {
      *size = chunk->len;
      return chunk->p;
    }
  }
  *size = 0;
  return NULL;
}

#if 0
/* interlace */
{
  int xp, yp, wp, hp;		/* x,y, width,height of pass */
  int xs = 8, ys = 8;		/* x,y step */

  /* pass珉WvZ */
  /* (0,0), (4,0), (0,4), (2,0), (0,2), (1,0), (0,1) */
  xp = (8 >> (pass / 2)) & 7;
  yp = 0;
  if (pass & 1) {
    yp = xp;
    xp = 0;
  }
  /* passщzvZ */
  /* (8,8), (8,8), (4,8), (4,4), (2,4), (2,2), (1,2) */
  if (pass > 3) {
    xs >>= (pass - 1) / 2;
    ys >>= (pass - 2) / 2;
  }

  /* passpass̉摜̑傫vZ */
  wp = width + xs - (x + 1);
  hp = height + ys - (y + 1);
  if (wp > 0 && hp > 0) {
    wp /= xs;
    hp /= ys;
  } else {
    wp = 0;
    hp = 0;			/* passȂ̂ŎpassōČvZ */
  }
}
#endif

/* *png->prev; filterꂽO̍sf[^
 * *png->line; filterꂽf[^
 * png->filterMode; ̍s̃tB^[h
 * *png->true; ̍sgD[J[(32bpp)ɂ́B
 * s܂scaling (ƌF)B
 * sƂɏ͍̂̂߁B
 * ň̏ꍇ͈s3000sNZ炢EB
 */

/** byteWriterf[^󂯎Đ`ďo */
void Png_putByte(Png_t png, byte b)
{
  Png_putByteString(png, &b, 1); /* 蔲 */
}

void Png_putByteString(Png_t png, byte *bp, int len)
{
  int x = png->x;
  int wp = png->scan_width;	/* width/height of pass */
  int hp = png->height;
  int bpp = (png->bpp + 7) / 8;
  int xp = 0, yp = 0;		/* initial x/y of pass */
  int xs = 1, ys = 1;		/* x,y step */

 newPass:
  if (png->interlace) {
    while (png->pass < 8) {
      /* pass珉WvZ */
      /* (0,0), (4,0), (0,4), (2,0), (0,2), (1,0), (0,1) */
      xp = (8 >> (png->pass / 2)) & 7;
      if (png->pass & 1) {
	yp = xp;
	xp = 0;
      }
      /* passщzvZ */
      /* (8,8), (8,8), (4,8), (4,4), (2,4), (2,2), (1,2) */
      xs = ys = 8;
      if (png->pass >= 3) {
	xs >>= (png->pass - 1) / 2;
	ys >>= (png->pass - 2) / 2;
      }

      /* passpass̉摜̑傫vZ */
      wp = png->width + xs - (xp + 1);
      hp = png->height + ys - (yp + 1);
      if (wp < 0 || hp < 0) {
	wp = 0;
	hp = 0;
	png->pass++;		/* passȂ̂ŎpassōČvZ */
      } else {
	wp /= xs;		/* pixelP */
	hp /= ys;
	wp = (wp * png->bpp + 7) / 8; /* byteP */
	break;
      }
    }

    if (png->y < 0) {
      png->y = 0;
    }
  }

  while (len > 0) {
    byte *line = png->line;
    byte *prev = png->prev;

    if (x < 0) {		/* Vs */
      len--;
      png->filterMode = *bp++;
      x = 0;
      continue;
    } else {
      int limit = wp < x + len ? wp - x : len;
      len -= limit;
      if (png->filterMode < 2) {
	if (png->filterMode == 1) { /* Sub */
	  for (; limit > 0; bp++, x++, limit--) {
	    byte left = x >= bpp ? line[x - bpp] : 0;
	    line[x] = left + *bp;
	  }
	} else {			/* None */
	none:
	  for (; limit > 0; bp++, x++, limit--)
	    line[x] = *bp;
	}
      } else if (png->filterMode == 2) {/* Up */
	if (png->y == 0)
	  goto none;
	for (; limit > 0; bp++, x++, limit--)
	  line[x] = prev[x] + *bp;
      } else if (png->filterMode < 4) { /* Average */
	for (; limit > 0; bp++, x++, limit--) {
	  byte left = (x >= bpp) ? line[x - bpp] : 0;
	  line[x] = (prev[x] + left) / 2 + *bp;
	}
      } else if (png->filterMode == 4) { /* paeth */
	for (; limit > 0; bp++, x++, limit--) {
	  byte left, upleft;
	  if (x >= bpp) {
	    left = line[x - bpp];
	    upleft = prev[x - bpp];
	  } else
	    left = upleft = 0;
	  line[x] = paeth_predictor(left, prev[x], upleft) + *bp;
	}
      }
    }
    if (x >= wp) {		/* s */
      scaling(png, xs,ys, xp,yp); /* png->truescaling */
      png->line = prev;
      png->prev = line;
      x = -1;
      if (++png->y >= hp) {
	png->y = -1;		/* passI */
	png->pass++;
	goto newPass;
      }
    }
  }
  png->x = x;
}

/** bitXg[`(8rbgȉTvp) */
/* ӏłĂ΂ȂinlineɂĂ܂
 * G[ԂȂ͈̓`FbNȂ */
static inline void BitBuffer_init(BitBuffer_t buf, byte *p)
{
  buf->p = p;
  buf->cnt = 0;
}

/* big endian */
static inline uint BitBuffer_getBits(BitBuffer_t buf, int len)
{
  uint x = 0;
  if (len == 8 && buf->cnt == 0) /* 8bit optimized */
    return *buf->p++;

  while (len > 0) {
    if (buf->cnt == 0) {
      buf->tmp = *buf->p++;
      buf->cnt = 8;
    }
    if (len < buf->cnt) {
      x = (x << len) | (buf->tmp >> (buf->cnt - len));
      buf->cnt -= len;
      buf->tmp &= 0xff >> (8 - buf->cnt);
      break;			/* len = 0 */
    } else {
      x = (x << buf->cnt) | buf->tmp;
      len -= buf->cnt;
      buf->cnt = 0;
    }
  }
  return x;
}

/* J[^Cv0(O[XP[)̃rbg[x1,2,4̓pbg
 * CfbNXJ[ɉ񂷁Brbg[x16͐؂̂Ă8rbgɂB
 */
static inline byte alphaBlend(byte x, byte y, byte a)
{				/* Ƃ x = 0xff, a = 0xff ̂Ƃ */
  int c;
  c = x * a + y * (255 - a);	/* c = 0xfe01 */
  c += c >> 8;			/* c = 0xfeff */
  c += 128;			/* c = 0xff7f */
  c >>= 8;			/* c = 0xff */
  return c;
}

/* sgD[J[ɕϊ, Fɉ */
static void scaling(Png_t png, int xs, int ys, int xp, int yp)
{
  byte *src = png->line;
  byte *img = png->true;
  int i = png->width;

  if (png->interlace) {
    img = &png->image[(png->width * (png->y * ys + yp) + xp) * 4];
    i = (i + xs - xp - 1) / xs;
  }
  xs *= 4;

  if (~png->clr_type & PALETTE_USED) {
    const int step = (png->bit_depth == 8) ? 1 : 2;
    if (~png->clr_type & TRUECOLOR_USED) { /* grey scale */
      if (png->bit_depth < 8)	/* 1,2,4 bit grey scale */
	goto indexed;
      else if (~png->clr_type & ALPHA_USED) { /* 8/16 bit grey scale */
	for (; i; img += xs, src += step, i--) {
	  img[0] = img[1] = img[2] = *src;
	  img[3] = 0;
	}
      } else {			/* 8/16bit gray scale + alpha */
	for (; i; img += xs, src += step * 2, i--) {
	  int alpha = src[step];
	  img[0] = img[1] = img[2] = 
	    (alpha == 0) ? 255 :
	    (alpha == 255) ? *src :
	    alphaBlend(*src, 255, alpha);
	  img[3] = 0;
	}
      }
    } else {
      if (~png->clr_type & ALPHA_USED) { /* true color */
	for (; i; img += xs, src += step * 3, i--) {
	  img[0] = src[step*2];
	  img[1] = src[step];
	  img[2] = src[0];
	  img[3] = 0;
	}
      } else {			/* true color + alpha */
     	for (; i; img += xs, src += step * 4, i--) {
	  int alpha = src[step * 3];
	  if (alpha == 0) {
	    img[0] = img[1] = img[2] = 255;
	  } else if (alpha == 255) {
	    img[0] = src[step * 2];
	    img[1] = src[step];
	    img[2] = src[0];
	  } else {
	    img[0] = alphaBlend(src[step * 2], 255, alpha);
	    img[1] = alphaBlend(src[step], 255, alpha);
	    img[2] = alphaBlend(src[0], 255, alpha);
	  }
	  img[3] = 0;
	}
      }
    }
  } else {
    BitBuffer buf[1];
  indexed:
    BitBuffer_init(buf, src);
    for (; i; i--, img += xs) {
      int index = BitBuffer_getBits(buf, png->bit_depth);
      img[0] = png->plte[index].b;
      img[1] = png->plte[index].g;
      img[2] = png->plte[index].r;
      img[3] = 0;
    }
  }
  if (png->gosakaku) {
    (*png->gosakaku)(0, (int*)&png->true[png->width * 4], png->width, (int*)png->true, png->image);
    png->image += png->lineLength;
  } else {
    png->true += png->lineLength;
  }
}
